"""Timeseries Class"""
from .index import Indexed
from . import general_utils


class Timeseries(Indexed):
    """Parent class for Tidegauge and other timeseries type datasets
    Common methods ...
    """

    pass
